/*
 *  Copyright (C) 2004 Christos Tsantilas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "c-icap.h"
#include "service.h"
#include "header.h"
#include "body.h"
#include "simple_api.h"
#include "debug.h"
#include "ci_threads.h"
#include <memcache.h>
#include <errno.h>
#undef HAVE_CONFIG_H
#include <odbx.h>
#include <regex.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h>
#ifndef REG_BASIC
#define REG_BASIC 0
#endif

#define CACHETTL	3600

struct memcache *mc;

int    url_filter_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf);
int    url_filter_post_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf);
void   url_filter_close_service (ci_service_xdata_t *srv_xdata);
void * url_filter_init_request_data(ci_request_t *req);
void   url_filter_release_data(void *data);
int    url_filter_process(ci_request_t *);
int    url_filter_check_preview(char *preview_data,int preview_data_len,ci_request_t *);
int    url_filter_io(char *wbuf,int *wlen,char *rbuf,int *rlen,int iseof,ci_request_t *req);
//int    url_filter_write(char *buf,int len ,int iseof,request_t *req);
//int    url_filter_read(char *buf,int len,request_t *req);
static unsigned dotip2int (char *str);
static unsigned str2int (const char *str);

static char	*MemcachedServers;
static char	*Host2CatServer;
static char	*DBEngine;
static char	*DBHost;
static char	*DBPort;
static char	*DBName;
static char	*DBUser;
static char	*DBPass;
static char	*RedirectUrl;
static char	*ClientIPHeader = "X-Client-IP";
static char	*AuthUserHeader = "X-Authenticated-User";
static char	*AuthUserPrefix;
static int	AuthUserEncoded = 1;
// Values 0 - ACCEPT, 1 - REJECT, 2 - REDIRECT
static int	DBOpenErrorPolicy = 1;	// SQLite database doesn't opened
static int	DBQueryErrorPolicy = 0;	// Error executing SQLite query
static int	UnknownRolePolicy = 1;		// Can't find role for user@ip
static int	MemcachedConnectErrorPolicy = 1;// No connection to memcached server(s)
static int	MemcachedMissPolicy = 0;	// Host category missed at memcached server(s)
static int	EmptyCategoryPolicy = 0;	// No categories for host
static int	CategoryNotFoundPolicy = 1;	// Host categories doesn't exists in roles definition
static int	SyslogPriority = LOG_INFO;	// Priority for syslog()

static int	fd = -1;
static struct sockaddr_in	addr;
static odbx_t	*dbh;
static char	*req_type_mod;			// Type of request (RESPMOD OR REQMOD)

static ci_thread_mutex_t mtx;

static struct ci_conf_entry conf_variables[]={
     {"MemcachedServers",&MemcachedServers,ci_cfg_set_str,NULL},
     {"Host2CatServer",&Host2CatServer,ci_cfg_set_str,NULL},
     {"DBEngine",&DBEngine,ci_cfg_set_str,NULL},
     {"DBHost",&DBHost,ci_cfg_set_str,NULL},
     {"DBPort",&DBPort,ci_cfg_set_str,NULL},
     {"DBName",&DBName,ci_cfg_set_str,NULL},
     {"DBUser",&DBUser,ci_cfg_set_str,NULL},
     {"DBPass",&DBPass,ci_cfg_set_str,NULL},
     {"RedirectUrl",&RedirectUrl,ci_cfg_set_str,NULL},
     {"ClientIPHeader",&ClientIPHeader,ci_cfg_set_str,NULL},
     {"AuthUserHeader",&AuthUserHeader,ci_cfg_set_str,NULL},
     {"AuthUserPrefix",&AuthUserPrefix,ci_cfg_set_str,NULL},
     {"AuthUserEncoded",&AuthUserEncoded,ci_cfg_set_int,NULL},
     {"DBOpenErrorPolicy",&DBOpenErrorPolicy,ci_cfg_set_int,NULL},
     {"DBQueryErrorPolicy",&DBQueryErrorPolicy,ci_cfg_set_int,NULL},
     {"UnknownRolePolicy",&UnknownRolePolicy,ci_cfg_set_int,NULL},
     {"MemcachedConnectErrorPolicy",&MemcachedConnectErrorPolicy,ci_cfg_set_int,NULL},
     {"MemcachedMissPolicy",&MemcachedMissPolicy,ci_cfg_set_int,NULL},
     {"EmptyCategoryPolicy",&EmptyCategoryPolicy,ci_cfg_set_int,NULL},
     {"CategoryNotFoundPolicy",&CategoryNotFoundPolicy,ci_cfg_set_int,NULL},
     {"SyslogPriority",&SyslogPriority,ci_cfg_set_int,NULL},
     {NULL,NULL,NULL,NULL}
};


//service_module echo={
CI_DECLARE_MOD_DATA ci_service_module_t service={
     "url_filter",
     "Url Filter service",
     ICAP_RESPMOD | ICAP_REQMOD,
     url_filter_init_service, /* init_service*/
     url_filter_post_init_service, /*post_init_service*/
     url_filter_close_service, /*close_Service*/
     url_filter_init_request_data,/* init_request_data*/
     url_filter_release_data, /*Release request data*/
     url_filter_check_preview,
     url_filter_process,
     url_filter_io,
     conf_variables,
     NULL
};
 
struct url_filter_data{
     ci_cached_file_t *body;
     int denied;
};

enum http_methods {HTTP_UNKNOWN=0,HTTP_GET, HTTP_POST};

struct http_info{
    int http_major;
    int http_minor;
    int method;
    char site[CI_MAXHOSTNAMELEN+1];
    char page[1024]; /*I think it is enough*/
//    char server_ip [CI_IPLEN + 1];
    char *client_ip;
    char auth_user [MAX_USERNAME_LEN + 1];
};


int url_filter_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf){
    unsigned int xops;
    printf("Initialization of url_filter module......\n");
    ci_service_set_preview(srv_xdata, 0);
    xops = CI_XCLIENTIP | CI_XSERVERIP;
    xops |= CI_XAUTHENTICATEDUSER | CI_XAUTHENTICATEDGROUPS;
    ci_service_set_xopts(srv_xdata, xops);
    return CI_OK;
}

int url_filter_post_init_service (ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf)
{
	int	nservers, retval;
	char	*p, *p1, *ep;

	if (!MemcachedServers || !*MemcachedServers) {
		ci_debug_printf (1, "Empty MemcachedServers list\n");
		return CI_ERROR;
	}

	if (!(mc = mc_new ())) {
		ci_debug_printf (1, "No memory for mc structure\n");
		return CI_ERROR;
	}

	ep = MemcachedServers + strlen(MemcachedServers);
	for (nservers = 0, p = MemcachedServers; p < ep; p = p1 + 1) {
		// p points to beginning of server, p1 points to end of server
		p1 = strchr (p, ',');
		if (!p1)        p1 = p + strlen (p);

		if (p1 == p) {
			ci_debug_printf (5, "Empty server in '%s' - ignored\n", MemcachedServers);
		} else if ((retval = mc_server_add5 (mc, p, p1 - p))) {
			ci_debug_printf (5, "server add for server %.*s returns %d\n", (int)(p1 - p), p, retval);
		} else {
			nservers++;
			ci_debug_printf (5, "server %.*s successfully added\n", (int)(p1 - p), p);
		}
	}

	if (!nservers) {
		ci_debug_printf (1, "No servers added\n");
		mc_free (mc);
		mc = NULL;
		return CI_ERROR;
	}

	if (!Host2CatServer || !*Host2CatServer) {
		ci_debug_printf (1, "Empty Host2CatServer\n");
		return CI_ERROR;
	}

	if (!(p = strchr (Host2CatServer, ':'))) {
		ci_debug_printf (1, "No port number in Host2CatServer %s\n", Host2CatServer);
		return CI_ERROR;
	}

	*p = 0;
	bzero (&addr, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_port = htons (atoi (p + 1));
	inet_pton (AF_INET, Host2CatServer, &addr.sin_addr);
	*p = ':';
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd == -1) {
		ci_debug_printf (1, "Can't open UDP socket to Host2CatServer: %s\n", strerror (errno));
		return CI_ERROR;
	}

	if (!DBEngine || !*DBEngine) {
		ci_debug_printf (1, "Empty DBEngine\n");
		return CI_ERROR;
	}

	retval = odbx_init (&dbh, DBEngine, DBHost, DBPort);
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error initializing DB handle: %s\n", odbx_error (dbh, retval));
		return CI_ERROR;
	}

	if (!DBName || !*DBName) {
		ci_debug_printf (1, "Empty DBName\n");
		return CI_ERROR;
	}

	retval = odbx_bind_simple (dbh, DBName, DBUser, DBPass);
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error binding to database: %s\n", odbx_error (dbh, retval));
		odbx_finish (dbh);
		dbh = NULL;
		return CI_ERROR;
	}

	if (!RedirectUrl || !*RedirectUrl)	ci_debug_printf (1, "Empty redirect URL\n");

	if (ci_thread_mutex_init(&mtx))
		return CI_ERROR;

	return CI_OK;
}


void url_filter_close_service (ci_service_xdata_t *srv_xdata)
{
	if (mc) {
		mc_free (mc);
		mc = NULL;
	}
	if (fd != -1)	close (fd);
	if (dbh) {
		odbx_unbind (dbh);
		odbx_finish (dbh);
	}

	ci_thread_mutex_destroy(&mtx);
}

void *url_filter_init_request_data(ci_request_t *req){
    struct url_filter_data *uc=malloc(sizeof(struct url_filter_data));
    uc->body=NULL;
    uc->denied = 0;
    return uc; /*Get from a pool of pre-allocated structs better......*/
}


void url_filter_release_data(void *data){
    struct url_filter_data *uc=data;
    if(uc->body)
	ci_cached_file_destroy(uc->body);
    free(uc); /*Return object to pool.....*/
}


static int get_http_info(ci_request_t *req,ci_headers_list_t *req_header , struct http_info *httpinf){
    char *str;
    int i;

        /*Now get the site name*/
    str=ci_headers_value(req_header,"Host");
    strncpy(httpinf->site,str,CI_MAXHOSTNAMELEN);
    httpinf->site[CI_MAXHOSTNAMELEN]='\0';

    str=req_header->headers[0];
    if(str[0]=='g' || str[0]=='G') /*Get request....*/
	httpinf->method=HTTP_GET;
    else if(str[0]=='p' || str[0]=='P') /*post request....*/
	httpinf->method=HTTP_POST;
    else{
	 httpinf->method=HTTP_UNKNOWN;
	return 0;
    }
    if((str=strchr(str,' '))==NULL){ /*The request must have the form:GETPOST page HTTP/X.X */
	 return 0;
    }
    while(*str==' ') str++;
    i=0;
    while(*str!=' ' && *str!='\0' && i<1022) /*copy page to the struct.*/
	httpinf->page[i++]=*str++;
    httpinf->page[i]='\0';

    if(*str!=' '){ /*Where is the protocol info?????*/
	 return 0;
    }
    while(*str==' ') str++;
    if(*str!='H' || *(str+4)!='/'){ /*Not in HTTP/X.X form*/
	 return 0;
    }
    str+=5;
    httpinf->http_major=strtol(str,&str,10);
    if(*str!='.'){
	 return 0;
    }
    str++;
    httpinf->http_minor=strtol(str,&str,10);

    
    return 1;
}

static int check_destination (struct http_info *httpinf, int policy)
{
	odbx_result_t	*sth;
	unsigned	ip, role = 0, *ncats;
	int	i, rc = CategoryNotFoundPolicy, retval;
	char	*cats = NULL, *p, *p1, buf [1024], rname[64], *uname;
	char	key[64], *val = NULL;
	int	valsz, found;

	uname = AuthUserPrefix && *AuthUserPrefix && !strncmp (httpinf->auth_user, AuthUserPrefix, strlen (AuthUserPrefix)) ? httpinf->auth_user + strlen (AuthUserPrefix) : httpinf->auth_user;
	ip = dotip2int (httpinf->client_ip);
	snprintf (key, sizeof key, "%s@%s", uname, httpinf->client_ip);
	ci_debug_printf (5, "Uname: %s\n", uname);
	ci_debug_printf (5, "IP: %s\n", httpinf->client_ip);
	ci_debug_printf (5, "URL: %s\n", httpinf->page);
	ci_debug_printf (5, "Host: %s\n", httpinf->site);

	if (!dbh) {
		ci_debug_printf (1, "DB handle doesn't opened\n");
		rc = DBOpenErrorPolicy;
		goto print_log;
	}

	// Find user role
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	
	if (val) {
		char *ptr;
             
		ci_debug_printf(5, "Found role in cache\n");
		role = strtol(val, NULL, 10);
		ptr = strchr(val, ':') + 1;
		strlcpy(rname, ptr, sizeof rname);
		free(val);
		goto cached_role;
	}

	snprintf (buf, sizeof buf, "SELECT u.addr, u.mask, u.rolid, r.rname FROM users u, roles r WHERE u.uname = '%s' AND u.rolid = r.rolid ORDER BY mask DESC", uname);

	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}
	
	while ((retval = odbx_row_fetch (sth)) != 0) {
		unsigned	net, mask;
		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s', :%s\n", buf, odbx_error (dbh, retval));
			break;
		}
		net = str2int (odbx_field_value (sth, 0));
		mask = str2int (odbx_field_value (sth, 1));
		if ((ip & mask) == net) {
			role = str2int (odbx_field_value (sth, 2));
			strlcpy (rname, odbx_field_value (sth, 3), sizeof rname);
			break;
		}
	}

	odbx_result_finish(sth);
	ci_thread_mutex_unlock(&mtx); 

	// Find default role
	if (!role) {
		snprintf (buf, sizeof buf, "SELECT u.addr, u.mask, u.rolid, r.rname FROM users u, roles r WHERE u.uname = '%s' AND u.rolid = r.rolid ORDER BY mask DESC", "");

		ci_thread_mutex_lock(&mtx);
		retval = odbx_query (dbh, buf, strlen (buf));
		if (retval != ODBX_ERR_SUCCESS) {
			ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
			rc = DBQueryErrorPolicy;
			ci_thread_mutex_unlock(&mtx);
			goto print_log;
		}

		retval = odbx_result (dbh, &sth, NULL, 0);
		if (retval < 0) {
			ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
			rc = DBQueryErrorPolicy;
			ci_thread_mutex_unlock(&mtx);
			goto print_log;
		}

		while ((retval = odbx_row_fetch (sth)) != 0) {
			unsigned	net, mask;
			if (retval < 0) {
				ci_debug_printf (1, "Error fetching result row '%s', :%s\n", buf, odbx_error (dbh, retval));
				break;
			}
			net = str2int (odbx_field_value (sth, 0));
			mask = str2int (odbx_field_value (sth, 1));
			if ((ip & mask) == net) {
				role = str2int (odbx_field_value (sth, 2));
				strlcpy (rname, odbx_field_value (sth, 3), sizeof rname);
				break;
			}
		}
		odbx_result_finish(sth);
	        ci_thread_mutex_unlock(&mtx);	
	}
        
	if (role) {
		snprintf(buf, sizeof buf, "%d:%s", role, rname);
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), buf, strlen(buf) + 1, CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
	}

cached_role:
	ci_debug_printf (5, "Role: %s\n", role ? rname : "UNKNOWN");
	if (!role) {
		rc = UnknownRolePolicy;
		goto print_log;
	}

	// Read URL list
	snprintf (key, sizeof key, "url:%s", rname);
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	if (val) {
		char *s;
		ci_debug_printf(5, "Found URL list in cache\n");

		for (s = val; *s;) {
			regex_t	preg;
			const char *url;
			int action;

			action = strtol(s, NULL, 10);
			url = strchr(s, ':') + 1;
			s = strchr(url, ' ');
			*s++ = '\0';
		
			ci_debug_printf(9, "checking url %s against list: "
			    "action %d, regex %s", httpinf->page, action, url);
			rc = regcomp(&preg, url, REG_BASIC | REG_ICASE |
			    REG_NOSUB);
			if (rc) {
				ci_debug_printf(1, "Error compiling regular "
				    "expression '%s': %d\n", url, rc);
				continue;
			}
			rc = regexec(&preg, httpinf->page, 0, NULL, 0);
			regfree(&preg);
			if (!rc) {
				free(val);
				rc = action;
				goto print_log;
			}
		}
		free(val);
		goto cached_urls;
	}

	snprintf (buf, sizeof buf, "SELECT l.url, l.action FROM lists l, hierarchy h WHERE h.child = %d AND h.parent = l.rolid ORDER BY action DESC, rolid, stars, length (l.url)", role);

	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	// Check URL list
	val = NULL;
	valsz = 0;
	found = 0;
	while ((retval = odbx_row_fetch (sth)) != 0) {
		regex_t	preg;
		const char	*url;
		int	action;

		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s': %s\n", buf, odbx_error (dbh, retval));
			break;
		}

		if (val != (void *)-1) {
			char *ptr;
			int len;

			len = strlen(odbx_field_value(sth, 0)) +
			    strlen(odbx_field_value(sth, 1)) + 3;
			if ((ptr = realloc(val, valsz + len)) == NULL) {
				free(val);
				val = (void *)-1;
			} else {
				val = ptr;
				asprintf(&ptr, "%s:%s ",
				    odbx_field_value(sth, 1),
				    odbx_field_value(sth, 0));
				if (ptr == NULL) {
					free(val);
					val = (void *)-1;
				} else {
				        if (valsz == 0) {
						memcpy(val,ptr,len);
					} else {
						strcat(val, ptr);
					}
					valsz += len;
				}
			}
			
		}

		url = odbx_field_value (sth, 0);
		action = str2int (odbx_field_value (sth, 1));
		ci_debug_printf (9, "checking url %s against list: action %d, regex %s", httpinf->page, action, url);
		rc = regcomp (&preg, url, REG_BASIC | REG_ICASE | REG_NOSUB);
		if (rc) {
			ci_debug_printf (1, "Error compiling regular expression '%s': %d\n", url, rc);
			continue;
		}
		rc = regexec (&preg, httpinf->page, 0, NULL, 0);
		regfree (&preg);
		if (!rc && !found) {
			found = 1;
			rc = action;
		}
	}
	odbx_result_finish (sth);
        ci_thread_mutex_unlock(&mtx);
       
	if (val && val != (void *)-1) {
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), val, strlen(val) + 1,
		    CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
		free(val);
	}

	if (found)
		goto print_log;

cached_urls:
	// Get host category
	if (!mc) {
		rc = MemcachedConnectErrorPolicy;
		goto print_log;
	}
	
	ci_thread_mutex_lock(&mtx);
	cats = mc_aget (mc, (char *) httpinf->site, strlen (httpinf->site));
	ci_thread_mutex_unlock(&mtx);
	ci_debug_printf (5, "Cats: %s\n", cats ? cats : "MISSED");
	if (!cats) {
		if (fd != -1) {
			sendto (fd, httpinf->site, strlen (httpinf->site) + 1, 0, (struct sockaddr *)&addr, sizeof addr);
			struct timeval timenow;
			if (req_type_mod == "RESPMOD") {
			    usleep(500000);
			    gettimeofday(&timenow, NULL);
			    syslog(SyslogPriority,"%s %u %u %s", "RESPMOD SEARCH", timenow.tv_sec, timenow.tv_usec, httpinf->page);
			    ci_thread_mutex_lock(&mtx);
			    cats = mc_aget (mc, (char *) httpinf->site, strlen (httpinf->site));
			    ci_thread_mutex_unlock(&mtx);
			    if (cats) goto continue_cached_urls;
			}
		} 
		rc = MemcachedMissPolicy;
		goto print_log;
	}
continue_cached_urls:
	if (!*cats) {
		rc = EmptyCategoryPolicy;
		goto print_log;
	}
	
	// Convert cats to array of unsigned
	for (i = 1, p = strchr (cats, ':'); p; p = strchr (p + 1, ':'), i++)
		;
	ncats = calloc (i + 1, sizeof (unsigned));
	for (i = 0, p = cats, p1 = strchr (p, ':'); *p; p1 = strchr (p, ':'), i++) {
		if (p1)	*p1 = 0;
		ncats [i] = str2int (p);
		if (p1) {
			*p1 = ':';
			p = p1 + 1;
		} else
			p += strlen (p);
	}
	ncats [i] = 0;
		
	// Read category list
	snprintf(key, sizeof key, "cat:%d", role);
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	if (val) {
		char *s;
		ci_debug_printf(5, "Found category list in cache\n");

		for (s = val; *s;) {
			char *catstr;
			int cat, action;
			unsigned *u;

			action = strtol(s, NULL, 10);
			catstr = strchr(s, ':') + 1;
			s = strchr(catstr, ' ');
			*s++ = '\0';

			cat = str2int(catstr);
			for (u = ncats; *u; u++) {
				ci_debug_printf(9, "checking host category %d "
				    "against REJECT and REDIRECT category %d\n",
				    *u, cat);
				if (cat == *u) {
					free(val);
					rc = action;
					goto print_log;
				}
			}
		}

		free(val);
		goto print_log;
	}
	
	snprintf (buf, sizeof buf, "SELECT p.cat, p.action FROM perms p, hierarchy h WHERE h.child = %d AND h.parent = p.rolid ORDER BY action DESC, rolid", role);
        
	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		free (ncats);
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		free (ncats);
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	// Check category list
	val = NULL;
	valsz = 0;
	found = 0;
	while ((retval = odbx_row_fetch (sth)) != 0) {
		int	cat, action;
		unsigned *u;

		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s': %s\n", buf, odbx_error (dbh, retval));
			break;
		}

		if (val != (void *)-1) {
			char *ptr;
			int len;

			len = strlen(odbx_field_value(sth, 0)) +
			    strlen(odbx_field_value(sth, 1)) + 3;
			if ((ptr = realloc(val, valsz + len)) == NULL) {
				free(val);
				val = (void *)-1;
			} else {
				val = ptr;
				asprintf(&ptr, "%s:%s ",
				    odbx_field_value(sth, 1),
				    odbx_field_value(sth, 0));
				if (ptr == NULL) {
					free(val);
					val = (void *)-1;
				} else {
				        if (valsz == 0) {
						memcpy(val, ptr, len);
					} else {
						strcat(val, ptr);
					}
					valsz += len;
				}
			}
		    	  
		}

		cat = str2int (odbx_field_value (sth, 0));
		action = str2int (odbx_field_value (sth, 1));
		for (u = ncats; *u; u++) {
			ci_debug_printf (9, "checking host category %d against REJECT and REDIRECT category %d\n", *u, cat);
			if (cat == *u && !found) {
				found = 1;
				rc = action;
			}
		}
	}
	free (ncats);
	odbx_result_finish (sth);
        ci_thread_mutex_unlock(&mtx);

	if (val && val != (void *)-1) {
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), val, strlen(val) + 1,
		    CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
		
		free(val);
	}

print_log:
	// uname, ip, url, role, cats, action
	syslog (SyslogPriority, "%s %s %s %s %s %s %s",
		req_type_mod,
		uname,
		httpinf->client_ip,
		httpinf->page,
		role ? rname : "-",
		cats ? cats : "-",
		rc == 0 ? "ACCEPT" : rc == 1 ? "REJECT" : "REDIRECT"
	);
	if (cats)	free (cats);
	return rc;
}

static char *reject_message="<H1>Permition deny!<H1>";
static char *redirect_message="<H1>You are redirected!<H1>";
//static char *server_ip = "X-Server-IP";

int url_filter_check_preview(char *preview_data,int preview_data_len, ci_request_t *req){
    ci_headers_list_t* req_header;
    struct url_filter_data *uc=ci_service_data(req);
    struct http_info httpinf;
    int rc;
    char* authuser = NULL;

    if (ci_req_type(req) != ICAP_RESPMOD)
	req_type_mod = "REQMOD";
    else 
	req_type_mod = "RESPMOD";

    if((req_header=ci_http_request_headers(req))==NULL) /*It is not possible but who knows .....*/
	return CI_ERROR;
	
//    httpinf.server_ip = ci_headers_value (req->request_header, server_ip);
//    ci_conn_remote_ip (req->connection, httpinf.server_ip);
    httpinf.client_ip = ci_headers_value (req->request_header, ClientIPHeader);
    httpinf.auth_user [0] = 0;

    if ( !(authuser = ci_headers_value (req->request_header, AuthUserHeader)) ) authuser = "";

    if (AuthUserEncoded)
    	ci_base64_decode (authuser, httpinf.auth_user, MAX_USERNAME_LEN);
    else
    	strncpy(httpinf.auth_user, authuser, MAX_USERNAME_LEN);

    get_http_info(req,req_header, &httpinf);

    ci_debug_printf(9,"URL  to host %s\n",httpinf.site);
    ci_debug_printf(9,"URL  page %s\n",httpinf.page);

    rc=check_destination(&httpinf, 1);
    ci_debug_printf (5, "Action: %s\n", rc == 0 ? "ACCEPT" : rc == 1 ? "REJECT" : "REDIRECT");

    if (!rc) {
	 /*if we are inside preview negotiation or client allow204 responces oudsite of preview then*/
	 if(preview_data || ci_req_allow204(req)) 
	      return CI_MOD_ALLOW204;
	 
	 /*
	   Squid does not support preview of data in reqmod requests neither 204 responces outside preview
	   so we need to read all the body if exists and send it back to squid.
	   Allocate a new body for it 
	 */
	 if(ci_req_hasbody(req)){
	      int clen=ci_http_content_length(req)+100;
	      uc->body=ci_cached_file_new(clen);
	 }

    } else if (rc == 2 && RedirectUrl && *RedirectUrl) {
	 char	buf [strlen (RedirectUrl) + 20];
	 /*The URL is not a good one so....*/
	 ci_debug_printf(9,"Oh!!! we are going to redirect this site.....\n");

	 uc->denied = 1;
	 uc->body=ci_cached_file_new(strlen(redirect_message)+10);
	 ci_http_response_create(req,1,1); /*Build the responce headers*/

	 ci_http_response_add_header(req,"HTTP/1.1 302 Found");/*Send an 302 Found http responce to web client*/
	 ci_http_response_add_header(req,"Server: C-ICAP");
	 snprintf (buf, sizeof buf, "Location: %s", RedirectUrl);
	 ci_http_response_add_header(req,buf);
	 ci_http_response_add_header(req,"Content-Type: text/html");
	 ci_http_response_add_header(req,"Content-Language: en");
	 
	 ci_cached_file_write(uc->body,redirect_message,strlen(redirect_message),1);
    } else {
	 /*The URL is not a good one so....*/
	 ci_debug_printf(9,"Oh!!! we are going to deny this site.....\n");

	 uc->denied = 1;
	 uc->body=ci_cached_file_new(strlen(reject_message)+10);
	 ci_http_response_create(req,1,1); /*Build the responce headers*/

	 ci_http_response_add_header(req,"HTTP/1.1 403 Forbidden");/*Send an 403 Forbidden http responce to web client*/
	 ci_http_response_add_header(req,"Server: C-ICAP");
	 ci_http_response_add_header(req,"Content-Type: text/html");
	 ci_http_response_add_header(req,"Content-Language: en");
	 
	 ci_cached_file_write(uc->body,reject_message,strlen(reject_message),1);
    }

    unlock_data(req);
    return CI_MOD_CONTINUE;
}


int url_filter_process(ci_request_t *req){

/*
	  printf("Buffer size=%d, Data size=%d\n ",
		 ((struct membuf *)b)->bufsize,((struct membuf *)b)->endpos);
*/  
//    printf("url_filter process\n");
     return CI_MOD_DONE;     
}

int url_filter_io(char *wbuf, int *wlen, char *rbuf, int *rlen, int iseof,
                  ci_request_t * req)
{
     int ret;
     struct url_filter_data *uc = ci_service_data(req);
     if (!uc->body)
          return CI_ERROR;

     ret = CI_OK;
     if (uc->denied == 0) {
          if (rbuf && rlen) {
               *rlen = ci_cached_file_write(uc->body, rbuf, *rlen, iseof);
               if (*rlen == CI_ERROR)
                    ret = CI_ERROR;
          }
          else if (iseof)
               ci_cached_file_write(uc->body, NULL, 0, iseof);
     }

     if (wbuf && wlen) {
          *wlen = ci_cached_file_read(uc->body, wbuf, *wlen);
          if (*wlen == CI_ERROR)
               ret = CI_ERROR;
     }

     return ret;
}

static unsigned dotip2int (char *str)
{
	unsigned	ip = 0, byte = 0;

	for (; *str; str++) {
		if (*str == '.') {
			ip <<= 8;
			ip += byte;
			byte = 0;
		} else {
			byte *= 10;
			byte += *str - '0';
		}
	}
	ip <<= 8;
	ip += byte;
	return ip;
}

static unsigned str2int (const char *str)
{
	unsigned	ip = 0;

	if (!str)	return 0;
	for (; *str; str++) {
		ip *= 10;
		ip += *str - '0';
	}
	return ip;
}
